# Copyright (c) Alibaba, Inc. and its affiliates.
from .cross_entropy_loss import CrossEntropyLoss
from .focal_loss import FocalLoss
from .iou_loss import GIoULoss, IoULoss
from .mse_loss import JointsMSELoss
from .pytorch_metric_learning import *
