# Copyright (c) Alibaba, Inc. and its affiliates.
from easycv.utils.registry import Registry

MODELS = Registry('model')
BACKBONES = Registry('backbone')
NECKS = Registry('neck')
HEADS = Registry('head')
LOSSES = Registry('loss')
